import numpy as np
import scipy as sp
from sklearn import dummy
from sklearn.ensemble import GradientBoostingRegressor, RandomForestRegressor, _gb_losses
from sklearn.linear_model import ElasticNet, Lasso, LinearRegression, Ridge
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor

from sklearn2json import csr
from sklearn2json.common_functions import deserialize_tree, serialize_tree


def serialize_linear_regressor(model):
    serialized_model = {
        "meta": "linear-regression",
        "coef_": model.coef_.tolist(),
        "intercept_": model.intercept_.tolist(),
        "singular_": model.singular_.tolist(),
        "rank_": model.rank_,
        "n_features_in_": model.n_features_in_,
        # "_residues": model._residues,
        "fit_intercept": model.fit_intercept,
        "params": model.get_params(),
    }

    return serialized_model


def deserialize_linear_regressor(model_dict):
    model = LinearRegression(**model_dict["params"])
    model.singular_ = np.array(model_dict["singular_"])
    model.coef_ = np.array(model_dict["coef_"])
    # model._residues = np.float64(model_dict["_residues"])
    model.intercept_ = np.float64(model_dict["intercept_"])
    model.rank_ = model_dict["rank_"]
    if model_dict.get("n_features_in_"):
        model.n_features_in_ = model_dict["n_features_in_"]
    model.fit_intercept = model_dict["fit_intercept"]
    return model


def serialize_lasso_regressor(model):
    serialized_model = {
        "meta": "lasso-regression",
        "coef_": model.coef_.tolist(),
        "n_features_in_": model.n_features_in_,
        "dual_gap_": model.dual_gap_,
        "params": model.get_params(),
    }

    if isinstance(model.n_iter_, int):
        serialized_model["n_iter_"] = model.n_iter_
    else:
        serialized_model["n_iter_"] = model.n_iter_.tolist()

    if isinstance(model.intercept_, float):
        serialized_model["intercept_"] = model.intercept_
    else:
        serialized_model["intercept_"] = model.intercept_.tolist()

    return serialized_model


def deserialize_lasso_regressor(model_dict):
    model = Lasso(**model_dict["params"])

    model.coef_ = np.array(model_dict["coef_"])

    if isinstance(model_dict["n_iter_"], list):
        model.n_iter_ = np.array(model_dict["n_iter_"])
    else:
        model.n_iter_ = int(model_dict["n_iter_"])

    if isinstance(model_dict["intercept_"], list):
        model.intercept_ = np.array(model_dict["intercept_"])
    else:
        model.intercept_ = float(model_dict["intercept_"])
    if model_dict.get("n_features_in_"):
        model.n_features_in_ = model_dict["n_features_in_"]
    model.dual_gap_ = np.float64(model_dict["dual_gap_"])
    return model


def serialize_elastic_regressor(model):
    serialized_model = {
        "meta": "elasticnet-regression",
        "coef_": model.coef_.tolist(),
        "alpha": model.alpha,
        "dual_gap_": model.dual_gap_,
        "n_features_in_": model.n_features_in_,
        "params": model.get_params(),
    }

    if isinstance(model.n_iter_, int):
        serialized_model["n_iter_"] = model.n_iter_
    else:
        serialized_model["n_iter_"] = model.n_iter_.tolist()

    if isinstance(model.intercept_, float):
        serialized_model["intercept_"] = model.intercept_
    else:
        serialized_model["intercept_"] = model.intercept_.tolist()

    return serialized_model


def deserialize_elastic_regressor(model_dict):
    model = ElasticNet(**model_dict["params"])
    if model_dict.get("n_features_in_"):
        model.n_features_in_ = model_dict["n_features_in_"]
    model.coef_ = np.array(model_dict["coef_"])
    model.alpha = float(model_dict["alpha"])
    model.dual_gap_ = np.float64(model_dict["dual_gap_"])
    if isinstance(model_dict["n_iter_"], list):
        model.n_iter_ = np.array(model_dict["n_iter_"])
    else:
        model.n_iter_ = int(model_dict["n_iter_"])

    if isinstance(model_dict["intercept_"], list):
        model.intercept_ = np.array(model_dict["intercept_"])
    else:
        model.intercept_ = float(model_dict["intercept_"])

    return model


def serialize_ridge_regressor(model):
    serialized_model = {
        "meta": "ridge-regression",
        "n_features_in_": model.n_features_in_,
        "coef_": model.coef_.tolist(),
        "n_iter_": model.n_iter_,
        "params": model.get_params(),
    }

    if model.n_iter_:
        serialized_model["n_iter_"] = model.n_iter_.tolist()

    if isinstance(model.intercept_, float):
        serialized_model["intercept_"] = model.intercept_
    else:
        serialized_model["intercept_"] = model.intercept_.tolist()

    return serialized_model


def deserialize_ridge_regressor(model_dict):
    model = Ridge(**model_dict["params"])
    if model_dict.get("n_features_in_"):
        model.n_features_in_ = model_dict["n_features_in_"]
    model.coef_ = np.array(model_dict["coef_"])

    if model_dict.get("n_iter_"):
        model.n_iter_ = np.array(model_dict["n_iter_"])
    else:
        model.n_iter_ = model_dict["n_iter_"]

    if isinstance(model_dict["intercept_"], list):
        model.intercept_ = np.array(model_dict["intercept_"])
    else:
        model.intercept_ = float(model_dict["intercept_"])

    return model


def serialize_svr(model):
    serialized_model = {
        "meta": "svr",
        "n_features_in_": model.n_features_in_,
        "support_": model.support_.tolist(),
        "fit_status_": model.fit_status_,
        "_n_support": model._n_support.tolist(),
        "intercept_": model.intercept_.tolist(),
        "_probA": model._probA.tolist(),
        "_probB": model._probB.tolist(),
        "_intercept_": model._intercept_.tolist(),
        "shape_fit_": model.shape_fit_,
        "_gamma": model._gamma,
        "params": model.get_params(),
        "n_iter_": model.n_iter_,
        "_num_iter": model._num_iter.tolist(),
    }

    if isinstance(model.support_vectors_, sp.sparse.csr_matrix):
        serialized_model["support_vectors_"] = csr.serialize_csr_matrix(model.support_vectors_)
    elif isinstance(model.support_vectors_, np.ndarray):
        serialized_model["support_vectors_"] = model.support_vectors_.tolist()

    if isinstance(model.dual_coef_, sp.sparse.csr_matrix):
        serialized_model["dual_coef_"] = csr.serialize_csr_matrix(model.dual_coef_)
    elif isinstance(model.dual_coef_, np.ndarray):
        serialized_model["dual_coef_"] = model.dual_coef_.tolist()

    if isinstance(model._dual_coef_, sp.sparse.csr_matrix):
        serialized_model["_dual_coef_"] = csr.serialize_csr_matrix(model._dual_coef_)
    elif isinstance(model._dual_coef_, np.ndarray):
        serialized_model["_dual_coef_"] = model._dual_coef_.tolist()

    return serialized_model


def deserialize_svr(model_dict):
    model = SVR(**model_dict["params"])
    model.fit_status_ = model_dict["fit_status_"]
    model.shape_fit_ = tuple(model_dict["shape_fit_"])
    model._gamma = model_dict["_gamma"]
    if model_dict.get("n_features_in_"):
        model.n_features_in_ = model_dict["n_features_in_"]
    model.support_ = np.array(model_dict["support_"]).astype(np.int32)
    model._n_support = np.array(model_dict["_n_support"]).astype(np.int32)
    model.intercept_ = np.array(model_dict["intercept_"]).astype(np.float64)
    model._probA = np.array(model_dict["_probA"]).astype(np.float64)
    model._probB = np.array(model_dict["_probB"]).astype(np.float64)
    model._intercept_ = np.array(model_dict["_intercept_"]).astype(np.float64)
    model.n_iter_ = model_dict["n_iter_"]
    model._num_iter = np.array(model_dict["_num_iter"]).astype(np.int32)

    if "meta" in model_dict["support_vectors_"] and model_dict["support_vectors_"]["meta"] == "csr":
        model.support_vectors_ = csr.deserialize_csr_matrix(model_dict["support_vectors_"])
        model._sparse = True
    else:
        model.support_vectors_ = np.array(model_dict["support_vectors_"]).astype(np.float64)
        model._sparse = False

    if "meta" in model_dict["dual_coef_"] and model_dict["dual_coef_"]["meta"] == "csr":
        model.dual_coef_ = csr.deserialize_csr_matrix(model_dict["dual_coef_"])
    else:
        model.dual_coef_ = np.array(model_dict["dual_coef_"]).astype(np.float64)

    if "meta" in model_dict["_dual_coef_"] and model_dict["_dual_coef_"]["meta"] == "csr":
        model._dual_coef_ = csr.deserialize_csr_matrix(model_dict["_dual_coef_"])
    else:
        model._dual_coef_ = np.array(model_dict["_dual_coef_"]).astype(np.float64)

    return model


def serialize_dummy_regressor(model):
    model.constant = model.constant_.tolist()
    return model.__dict__


def serialize_gradient_boosting_regressor(model):
    serialized_model = {
        "meta": "gb-regression",
        "max_features_": model.max_features_,
        # "n_features_": model.n_features_,
        "n_estimators_": model.n_estimators_,
        "n_features_in_": model.n_features_in_,
        "train_score_": model.train_score_.tolist(),
        "params": model.get_params(),
        "estimators_shape": list(model.estimators_.shape),
        "estimators_": [],
    }

    if isinstance(model.init_, dummy.DummyRegressor):
        serialized_model["init_"] = serialize_dummy_regressor(model.init_)
        serialized_model["init_"]["constant_"] = serialized_model["init_"]["constant_"].tolist()
        serialized_model["init_"]["meta"] = "dummy"
    elif isinstance(model.init_, str):
        serialized_model["init_"] = model.init_

    if isinstance(model._loss, _gb_losses.LeastSquaresError):
        serialized_model["_loss"] = "ls"
    elif isinstance(model._loss, _gb_losses.LeastAbsoluteError):
        serialized_model["_loss"] = "lad"
    elif isinstance(model._loss, _gb_losses.HuberLossFunction):
        serialized_model["_loss"] = "huber"
    elif isinstance(model._loss, _gb_losses.QuantileLossFunction):
        serialized_model["_loss"] = "quantile"

    if "priors" in model.init_.__dict__:
        serialized_model["priors"] = model.init_.priors.tolist()

    for tree in model.estimators_.reshape((-1,)):
        serialized_model["estimators_"].append(serialize_decision_tree_regressor(tree))
    return serialized_model


def deserialize_gradient_boosting_regressor(model_dict):
    model = GradientBoostingRegressor(**model_dict["params"])
    trees = [deserialize_decision_tree_regressor(tree) for tree in model_dict["estimators_"]]
    model.estimators_ = np.array(trees).reshape(model_dict["estimators_shape"])
    if "init_" in model_dict and model_dict["init_"]["meta"] == "dummy":
        model.init_ = dummy.DummyRegressor()
        model.init_.__dict__ = model_dict["init_"]
        model.init_.__dict__.pop("meta")
    model.n_features_in_ = model_dict["n_features_in_"]
    model.n_estimators_ = model_dict["n_estimators_"]
    model.train_score_ = np.array(model_dict["train_score_"])
    model.max_features_ = model_dict["max_features_"]
    if model_dict["_loss"] == "ls":
        model._loss = _gb_losses.LeastSquaresError()
    elif model_dict["_loss"] == "lad":
        model._loss = _gb_losses.LeastAbsoluteError()
    elif model_dict["_loss"] == "huber":
        model._loss = _gb_losses.HuberLossFunction(1)
    elif model_dict["_loss"] == "quantile":
        model._loss = _gb_losses.QuantileLossFunction(1)

    if "priors" in model_dict:
        model.init_.priors = np.array(model_dict["priors"])
    return model


def serialize_random_forest_regressor(model):
    serialized_model = {
        "meta": "rf-regression",
        "max_depth": model.max_depth,
        "min_samples_split": model.min_samples_split,
        "min_samples_leaf": model.min_samples_leaf,
        "min_weight_fraction_leaf": model.min_weight_fraction_leaf,
        "max_features": model.max_features,
        "max_leaf_nodes": model.max_leaf_nodes,
        "min_impurity_decrease": model.min_impurity_decrease,
        "n_features_in_": model.n_features_in_,
        "n_outputs_": model.n_outputs_,
        "estimators_": [serialize_decision_tree_regressor(decision_tree) for decision_tree in model.estimators_],
        "params": model.get_params(),
    }

    if "oob_score_" in model.__dict__:
        serialized_model["oob_score_"] = model.oob_score_
    if "oob_decision_function_" in model.__dict__:
        serialized_model["oob_prediction_"] = model.oob_prediction_.tolist()

    return serialized_model


def deserialize_random_forest_regressor(model_dict):
    model = RandomForestRegressor(**model_dict["params"])
    estimators = [deserialize_decision_tree_regressor(decision_tree) for decision_tree in model_dict["estimators_"]]
    model.estimators_ = estimators
    model.estimator_ = DecisionTreeRegressor()
    model.n_features_in_ = model_dict["n_features_in_"]
    model.n_outputs_ = model_dict["n_outputs_"]
    model.max_depth = model_dict["max_depth"]
    model.min_samples_split = model_dict["min_samples_split"]
    model.min_samples_leaf = model_dict["min_samples_leaf"]
    model.min_weight_fraction_leaf = model_dict["min_weight_fraction_leaf"]
    model.max_features = model_dict["max_features"]
    model.max_leaf_nodes = model_dict["max_leaf_nodes"]
    model.min_impurity_decrease = model_dict["min_impurity_decrease"]

    if "oob_score_" in model_dict:
        model.oob_score_ = model_dict["oob_score_"]
    if "oob_prediction_" in model_dict:
        model.oob_prediction_ = np.array(model_dict["oob_prediction_"])

    return model


def serialize_mlp_regressor(model):
    serialized_model = {
        "meta": "mlp-regression",
        "n_features_in_": model.n_features_in_,
        "coefs_": [elem.tolist() for elem in model.coefs_],
        "loss_": float(model.loss_),
        "intercepts_": [elem.tolist() for elem in model.intercepts_],
        "n_iter_": model.n_iter_,
        "n_layers_": model.n_layers_,
        "n_outputs_": model.n_outputs_,
        "out_activation_": model.out_activation_,
        "t_": model.t_,
        "best_loss_": float(model.best_loss_),
        "loss_curve_": model.loss_curve_,
        "_no_improvement_count": model._no_improvement_count,
        "params": model.get_params(),
        "validation_scores_": model.validation_scores_,
        "best_validation_score_": model.best_validation_score_,
    }

    return serialized_model


def deserialize_mlp_regressor(model_dict):
    model = MLPRegressor(**model_dict["params"])
    if model_dict.get("n_features_in_"):
        model.n_features_in_ = model_dict["n_features_in_"]
    model.coefs_ = [np.array(elem) for elem in model_dict["coefs_"]]
    model.loss_ = np.float64(model_dict["loss_"])
    model.intercepts_ = [np.array(elem) for elem in model_dict["intercepts_"]]
    model.n_iter_ = model_dict["n_iter_"]
    model.n_layers_ = model_dict["n_layers_"]
    model.n_outputs_ = model_dict["n_outputs_"]
    model.out_activation_ = model_dict["out_activation_"]
    model.hidden_layer_sizes = tuple(model.hidden_layer_sizes)
    model.t_ = model_dict["t_"]
    model._no_improvement_count = model_dict["_no_improvement_count"]
    model.best_loss_ = np.float64(model_dict["best_loss_"])
    model.loss_curve_ = model_dict["loss_curve_"]
    model.validation_scores_ = model_dict["validation_scores_"]
    model.best_validation_score_ = model_dict["best_validation_score_"]
    return model


def serialize_decision_tree_regressor(model):
    tree, dtypes = serialize_tree(model.tree_)
    serialized_model = {
        "meta": "decision-tree-regression",
        "feature_importances_": model.feature_importances_.tolist(),
        "max_features_": model.max_features_,
        "n_features_in_": model.n_features_in_,
        "n_outputs_": model.n_outputs_,
        "tree_": tree,
    }

    tree_dtypes = []
    for i in range(0, len(dtypes)):
        tree_dtypes.append(dtypes[i].str)

    serialized_model["tree_"]["nodes_dtype"] = tree_dtypes

    return serialized_model


def deserialize_decision_tree_regressor(model_dict):
    deserialized_decision_tree = DecisionTreeRegressor()

    deserialized_decision_tree.max_features_ = model_dict["max_features_"]
    deserialized_decision_tree.n_features_in_ = model_dict["n_features_in_"]
    deserialized_decision_tree.n_outputs_ = model_dict["n_outputs_"]
    tree = deserialize_tree(model_dict["tree_"], model_dict["n_features_in_"], 1, model_dict["n_outputs_"])
    deserialized_decision_tree.tree_ = tree

    return deserialized_decision_tree
